from django.db.models import ForeignKey
try:
    from django.core.exceptions import FieldDoesNotExist
except:
    # for old django version
    from django.db.models.fields import FieldDoesNotExist

def key_of_model(cls, *args, **kwargs):
    key_prefix = "%s.%s" % (cls.__module__, cls.__name__)

    if hasattr(cls, 'cache_version'):
        key_prefix += ".%s" % (getattr(cls, 'cache_version'))

    if args:
        if len(args) != 2:
            raise ValueError('args must be [field, val]')
        keys = "%s.%s" % (args[0], args[1])
    else:
        valid, msg = validate_fields(cls, kwargs)
        if not valid:
            raise ValueError(msg)
        keys = sorted(["%s.%s" % item for item in kwargs.items()])
        keys = '.'.join(keys)
    return "%s.%s" % (key_prefix, keys)


def key_of_model_list(cls, **kwargs):

    valid, msg = validate_fields(cls, kwargs)
    if not valid:
        raise ValueError(msg)

    key_prefix = "list.%s.%s" % (cls.__module__, cls.__name__)
    if hasattr(cls, 'list_cache_version'):
        key_prefix += ".%s" % (getattr(cls, 'list_cache_version'))

    # if len(kwargs) == 0 means i want get all
    if hasattr(cls, 'list_cache_version'):
        key_prefix += ".%s" % (getattr(cls, 'list_cache_version'))
    keys = sorted(["%s.%s" % item for item in kwargs.items()])
    keys = '.'.join(keys)
    return "%s.%s" % (key_prefix, keys)


def validate_fields(cls, fields):
    for key, value in fields.items():
        if key in ['pk', 'id']:
            continue
        if not key.endswith('_id') and '__' not in key:
            field = cls._meta.get_field(key)
            if isinstance(field, ForeignKey):
                return False, 'must use FIELD_id on related fields'
    return True, 'SUCCESS'



from django.core.cache import cache
from django.db.models.signals import post_save, post_delete


def apply_decorator(cls):
    def decorator(cls, func):
        def wrapper(*args, **kwargs):
            module_name = cls.__module__
            if module_name.endswith('.cache_keys'):
                module_name = module_name.rsplit('.', 1)[0]

            function_name = func.__name__[len('key_of_'):]

            return '.'.join([module_name, function_name,
                             str(func(*args, **kwargs) or '')])
        return wrapper

    for key, value in cls.__dict__.items():
        if key.startswith('key_of_'):
            setattr(cls, key, staticmethod(decorator(cls, value)))
    return cls


def clear_model_cache(Model, *args, **cache_kwargs):
    def clear_model(sender, instance, **kwargs):
        cache.clear_model(instance)
        if args:
            cache.clear_model(instance, *args)

        fields = cache_kwargs.get('model_list_fields')
        if fields is not None:
            cache.clear_model_list(instance, *fields)

    post_save.connect(clear_model, sender=Model, weak=False)
    post_delete.connect(clear_model, sender=Model, weak=False)



from django.db import models
from django.core.cache import cache
from django.utils.functional import curry


class CachedModel(models.Model):

    def _cached_FIELD(self, field, cache_exc):
        value = getattr(self, field.attname)
        Model = field.related_model
        return cache.get_model(Model, pk=value, cache_exc=cache_exc)

    class Meta:
        abstract = True


class ForeignKeyField(models.ForeignKey):

    def __init__(self, *args, **kwargs):
        self.cache_exc = kwargs.pop('cache_exc', False)
        super(ForeignKeyField, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name, **kwargs):
        super(models.ForeignKey, self).contribute_to_class(cls, name, **kwargs)
        if issubclass(cls, models.Model):
            setattr(cls, "cached_%s" % self.name,
                    property(curry(CachedModel._cached_FIELD, field=self,
                                   cache_exc=self.cache_exc)))


class OneToOneField(models.OneToOneField, ForeignKeyField):
    pass



from setuptools import setup

setup(
    name='django-cache-extension',
    version='1.1.8',
    description='cache extension for django',
    url='https://github.com/Beeblio/django-cache-extension',
    author='Shanbay python dev group',
    author_email='developer@shanbay.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    packages=['cache_extension'],
    package_data={'cache_extension': ['backends/*.py']},
    install_requires=['django-redis', 'redis', 'hiredis'],
)



from django.db import IntegrityError
from functools import partial
from cache_extension import cache_keys


class ModelNotExist(object):
    pass


class ExtensionCache(object):

    def get_attrs(self, model):
        return {
            f.attname: getattr(model, f.attname) for f in model._meta.fields}

    def get_or_create_model(self, cls, *args, **kwargs):
        try:
            model = self.get_model(cls, *args, **kwargs)
            return model, False
        except cls.DoesNotExist:
            if args:
                if len(args) > 1:
                    raise ValueError('multi field should pass by kwargs')
                kwargs = {'pk': args[0]}
            try:
                model = cls.objects.create(**kwargs)
            except IntegrityError:
                model = cls.objects.get(**kwargs)

            return model, True

    def get_model(self, cls, cache_exc=False, **kwargs):
        key = cache_keys.key_of_model(cls, **kwargs)
        attrs = self.get(key)
        if not attrs:
            try:
                model = cls.objects.get(**kwargs)
                self.set(key, self.get_attrs(model))
            except cls.DoesNotExist:
                if not cache_exc:
                    raise cls.DoesNotExist
                model = None
                self.set(key, ModelNotExist())
        elif isinstance(attrs, ModelNotExist):
            if not cache_exc:
                raise cls.DoesNotExist
            model = None
        else:
            fields = set([field.attname for field in cls._meta.fields])
            attrs_keys = set(attrs.keys())
            diff = attrs_keys - fields
            for key in diff:
                attrs.pop(key)
            model = cls(**attrs)
        return model

    def set_model(self, instance, *args):
        if instance is None:
            raise ValueError("can not cache a none model")
        key = self._make_model_key(instance, *args)
        attrs = self.get_attrs(instance)
        self.set(key, attrs)

    def clear_model(self, instance, *args):
        key = self._make_model_key(instance, *args)
        self.delete(key)

    def clear_models(self, instance, field, vals):
        keys = [cache_keys.key_of_model(
            instance, **{field: val}) for val in vals]
        self.delete_many(keys)

    def get_many_by_vals(self, vals, key_func):
        keys = [key_func(val) for val in vals]
        ret = self.get_many(keys)
        if ret:
            _ = {}
            m = dict(zip(keys, vals))
            for k, v in ret.items():
                _[m[k]] = v
            ret = _
        return ret

    def get_models(self, cls, vals, field='pk', version=None, sort=False):
        ''' get multiple models in multiple keys
        '''
        key_func = partial(cache_keys.key_of_model, cls, field)
        models = self.get_many_by_vals(vals, key_func=key_func)
        fields = set([f.attname for f in cls._meta.fields])
        for model in models.values():
            attrs_keys = set(model.keys())
            diff = attrs_keys - fields
            for key in diff:
                model.pop(key)

        exists_models = [cls(**m) for m in models.values()]
        if len(exists_models) == len(vals):
            models = exists_models
        else:
            miss_vals = set(vals) - set(models.keys())
            kwargs = {'%s__in' % field: miss_vals}
            miss_models = list(cls.objects.filter(**kwargs))
            data = {
                key_func(getattr(m, field)): self.get_attrs(m)
                for m in miss_models
            }
            self.set_many(data)
            models = exists_models + miss_models
        if sort:
            models = sorted(
                models, key=lambda m: vals.index(getattr(m, field)))
        return models

    def get_model_list(self, cls, **kwargs):
        ''' get multiple models in one key
        '''
        key = cache_keys.key_of_model_list(cls, **kwargs)
        models = self.get(key)
        if models:
            fields = set([f.attname for f in cls._meta.fields])
            for model in models:
                attrs_keys = set(model.keys())
                diff = attrs_keys - fields
                for key in diff:
                    model.pop(key)
            return [cls(**model) for model in models]
        models = list(cls.objects.filter(**kwargs))
        data = [self.get_attrs(m) for m in models]
        self.set(key, data)
        return models

    def set_model_list(self, cls, models=None, **kwargs):
        key = cache_keys.key_of_model_list(cls, **kwargs)
        if models is None:
            models = list(cls.objects.filter(**kwargs))
        data = [self.get_attrs(m) for m in models]
        self.set(key, data)

    def clear_model_list(self, instance, *args):
        key = self._make_model_list_key(instance, *args)
        self.delete(key)

    def _make_model_key(self, instance, *args):
        args = args or ['pk']
        kwargs = dict([(field, getattr(instance, field)) for field in args])
        return cache_keys.key_of_model(instance.__class__, **kwargs)

    def _make_model_list_key(self, instance, *args):
        kwargs = dict([(field, getattr(instance, field)) for field in args])
        key = cache_keys.key_of_model_list(instance.__class__, **kwargs)
        return key

    def fetch(self, key, block):
        data = self.get(key)
        if data is None:
            data = block()
            self.set(key, data)
        return data



from __future__ import absolute_import
from cache_extension.cache import ExtensionCache
from django_redis.cache import RedisCache, omit_exception
from django_redis.client.default import _main_exceptions, DefaultClient
from django_redis.exceptions import ConnectionInterrupted
from redis.client import StrictPipeline
from redis.exceptions import ResponseError

KEY_CMDS = ['exists', 'expire', 'expireat', 'rename', 'ttl']
# not support hstrlen
HASH_CMDS = [
    'hdel', 'hexists', 'hget', 'hgetall', 'hincrby', 'hincrbyfloat', 'hkeys',
    'hlen', 'hmget', 'hmset', 'hset', 'hsetnx', 'hvals', 'hscan',
]
# first arg must be key, so this cmds not support: brpoplpush, rpoplpush
LIST_CMDS = [
    'blpop', 'brpop', 'lindex', 'linsert', 'llen',
    'lpop', 'lpush', 'lpushx', 'lrange', 'lrem', 'lset',
    'ltrim', 'rpop', 'rpush', 'rpushx',
]
SET_CMDS = ['sadd', 'scard', 'sismember', 'smembers', 'spop', 'srem',
           'srandmember', 'sinter', 'sinterstore', 'sdiff', 'sdiffstore',
            'sunion', 'sunionstore']
ZSET_CMDS = [
    'zadd', 'zcard', 'zcount', 'zincrby', 'zrange', 'zrem',
    'zrevrange', 'zscore', 'zremrangebyrank'
]

SUPPORT_CMDS = KEY_CMDS + HASH_CMDS + LIST_CMDS + SET_CMDS + ZSET_CMDS


class ExtensionPipeline(StrictPipeline):
    def __init__(self, connection_pool, response_callbacks, transaction,
                 shard_hint, make_key_func):
        super(ExtensionPipeline, self).__init__(
            connection_pool,
            response_callbacks,
            transaction,
            shard_hint
        )

        self.make_key = make_key_func

    def pipeline_execute_command(self, *args, **options):
        args = list(args)
        key = args[1]
        args[1] = str(self.make_key(key))
        # print("pipeline exec", args)
        # print("pipeline exec opt", options)

        return super(
            ExtensionPipeline, self
        ).pipeline_execute_command(*args, **options)


class ExtensionClient(DefaultClient):

    def pipeline(self, transaction=True, shard_hint=None):
        client = self.get_client()

        return ExtensionPipeline(
            client.connection_pool,
            client.response_callbacks,
            transaction,
            shard_hint,
            self.make_key)

    def __getattr__(self, cmd):
        client = self.get_client()
        return getattr(client, cmd)


class ExtensionRedisBackend(ExtensionCache, RedisCache):

    def __init__(self, server, params):
        options = params.get("OPTIONS", {})
        default_client = "django_redis.client.DefaultClient"
        client_class = options.get('CLIENT_CLASS', default_client)
        options['CLIENT_CLASS'] = client_class
        params['options'] = options
        super(ExtensionRedisBackend, self).__init__(server, params)

    @omit_exception
    def incr(self, key, delta=1, version=None, client=None):

        if not client:
            client = self.client.get_client(write=True)

        key = self.make_key(key, version=version)

        try:
            try:
                value = client.incr(key, delta)
            except ResponseError:
                # if cached value or total value is greater than 64 bit signed
                # integer.
                # elif int is encoded. so redis sees the data as string.
                # In this situations redis will throw ResponseError

                # try to keep TTL of key
                timeout = client.ttl(key)
                value = self.get(key, version=version, client=client) + delta
                self.set(key, value, version=version, timeout=timeout,
                         client=client)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client, parent=e)

        return value

    def pipeline(self, transaction=True, shard_hint=None):
        return self.client.pipeline(transaction=transaction,
                                    shard_hint=shard_hint)

    def __getattr__(self, cmd):
        if cmd not in SUPPORT_CMDS:
            raise KeyError("not supported redis commands")

        func = getattr(self.client, cmd)

        def redis_cmd(key, *args, **kwargs):
            key = self.make_key(key)
            if cmd == 'rename':
                dest = args[0]
                dest = self.make_key(dest)
                args = list(args)
                args[0] = dest

            # print("exec", cmd, key, args, kwargs)
            return func(key, *args, **kwargs)

        return redis_cmd
